// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// Interface (both game and menu and general).

// fake keys for mouse support in menu
#define KEY_MOUSELEFT -1
#define KEY_MOUSERIGHT -2

CellType lm[MAXY][MAXX];
int lo[MAXY][MAXX];

void bufferTest();

bool needRestart = false;

int randColor() {
  return (rand() % 256) + ((rand() % 256) << 8) + ((rand() % 256) << 16);
  }

int randDarkColor() {
  switch(rand() % 4) {
    case 0: return (rand() % 128) + ((rand() % 128) << 8) + ((rand() % 128) << 16);
    case 1: return (rand() % 256);
    case 2: return (rand() % 256) << 8;
    case 3: return (rand() % 256) << 16;
    }
  }

int cmix(int a, int b, float f) {
  int res = 0;
  if(f < 0) f = 0;
  for(int k=0; k<4; k++) {
    int aa = (a >> (8*k)) & 255;
    int bb = (b >> (8*k)) & 255;
    int cc = int(aa + (bb-aa) * f + .5);
    res += cc << (8*k);
    }
  return res;
  }

char *qitoa(int i) {
  static char nbuf[16];
  sprintf(nbuf, "%d", i);
  return nbuf;
  }

void TDisplay::init() {
  if (SDL_Init(SDL_INIT_VIDEO) < 0) {
    printf("Failed to initialize SDL.\n");
    SDL_Quit();
    }

  if (Mix_OpenAudio(44100, MIX_DEFAULT_FORMAT, 2, 8192) != 0)
    printf("Failed to initialize audio.\n");
  else {
    Mix_AllocateChannels(16);
    sndBirth = Mix_LoadWAV("birth.wav");
    sndDeath = Mix_LoadWAV("death.wav");
    sndBlue = Mix_LoadWAV("blue.wav");
    sndGreen = Mix_LoadWAV("green.wav");
    sndRed = Mix_LoadWAV("red.wav");
    sndHit = Mix_LoadWAV("hit.wav");
    sndPlace = Mix_LoadWAV("place.wav");
    sndShoot = Mix_LoadWAV("shoot.wav");
    }

  
  if(TTF_Init() != 0) {
    printf("Failed to initialize TTF.\n");
    exit(2);
    }
  font = TTF_OpenFont("VeraBd.ttf", 16);
  if(!font) printf("Failed to read font.\n");

  stop  = false;
  
  setMode();
  }

void TDisplay::setMode() {

  sizeX = c.resX;
  sizeY = c.resY;
  stop  = false;
  
  surface = 
    SDL_SetVideoMode(sizeX, sizeY, 32, c.fs ? SDL_FULLSCREEN : 0);

  if(!surface) {
    sizeX = 800;
    sizeY = 600;
    surface = SDL_SetVideoMode(800, 600, 32, 0);
    }

  if(!surface) {
    printf("Failed to initialize SDL.\n");
    SDL_Quit();
    }
  SDL_WM_SetCaption("Untahris 3.1", 0);
  }

void TDisplay::playSoundAt(Mix_Chunk *sound, int x, int y) {
  if(sound) for(int i=0; i<16; i++) if(!Mix_Playing(i)) {
    Mix_PlayChannel(i, sound, 0);
    Mix_SetPanning(i, (255 * (g.sizeX-1-x)) / (g.sizeX-1), (255 * x) / (g.sizeX-1));
    return;
    }
  }

void TDisplay::loop() {
  SDL_LockSurface(surface);
  for(int y=0; y<sizeY; y++)
  for(int x=0; x<sizeX; x++) {
    int u = (x+2*y) ^ (y-2*x);
    putpix(x, y, 0x101 * (u & 0x3F));
    }
  SDL_UnlockSurface(surface);
  for(int y=0; y<g.sizeY; y++)
  for(int x=0; x<g.sizeX; x++) lm[y][x] = CellType(-1);
  
  while(!stop) {
    SDL_Event event;
    bool ev = SDL_PollEvent(&event);
    if(ev) switch (event.type) {

      case SDL_KEYDOWN:
        addKeypress(event.key.keysym.sym, true);
        break;

      case SDL_KEYUP:
        addKeypress(event.key.keysym.sym, false);
        break;
      
      case SDL_MOUSEBUTTONDOWN:
        addMouseEvent(event.button.x, event.button.y, 1, event.button.button);
        break;

      case SDL_MOUSEBUTTONUP:
        addMouseEvent(event.button.x, event.button.y, 2, event.button.button);
        break;

      case SDL_MOUSEMOTION:
        addMouseEvent(event.button.x, event.button.y, 0, 0);
        break;

      case SDL_QUIT:
        stop = true;
        return;
      }
    else {
      draw();
      dataTransfer();
      int t = getCurrentTime();
      g.upTime(t);
      SDL_Delay(1);
      }
    }
  }

void TDisplay::addKeypress(int key, bool type) {
  for(int i=0; i<size(lps); i++) {
    TLocalPlayer& l(lps[i]);
    for(int u=0; u<8; u++) if(l.keys[u] == key) {
      TKeypress k;
      k.time = getCurrentTime()+1;
      k.pressed = type;
      k.player = l.index;
      k.key = PlayerKey(u);
      addKeyToQueue(k);
      broadcastKey(k);
      }
    }
  
  int kid = 0;
  if(key == SDLK_F2) kid = 1;
  if(key == SDLK_F4) kid = 2;
  if(key == SDLK_F7) kid = 3;
  if(key == SDLK_F8) kid = 4;
  if(kid && type) {
    TKeypress k;
    k.time = getCurrentTime()+1;
    k.player = 0;
    k.key = kid;
    addKeyToQueue(k);
    broadcastKey(k);
    }
  
  if(key == 27) stop = true;
  }

void TDisplay::addMouseEvent(int x, int y, int type, int button) {
  static int mouser = 0;

  for(int i=0; i<size(lps); i++) {
    mouser %= size(lps);
    TLocalPlayer& l(lps[mouser]);
    if(g.pl[l.index]->typeval == -1) break;
    
    mouser++; mouser %= size(lps);
    }

  TLocalPlayer& l(lps[mouser]);
  
  if(x < ofsX || y <ofsY) return;
  x -= ofsX; x /= Res;
  y -= ofsY; y /= Res;
  if(x >= g.sizeX || y >= g.sizeY) return;
  
  TKeypress k;
  k.player = l.index;
  k.time = getCurrentTime()+1;
  k.pressed = type != 2;
  k.mx = x; k.my = y;
  k.key = type != 0 ? (button == 1 ? keyLMB : keyRMB) : keyMoveMouse;

  addKeyToQueue(k);
  broadcastKey(k);
  }

void TDisplay::putpix(int x, int y, int c) {
  if(x<0 || y<0 || x>=sizeX || y>=sizeY) return;
  
  if(x >= ofsX && y >= ofsY) {
    int zx = (x-ofsX) / d.Res;
    int zy = (y-ofsY) / d.Res;
    if(zx<g.sizeX && zy<g.sizeY) lm[zy][zx] = CellType(-1);
    }
  
  unsigned char *dst = (unsigned char*) surface->pixels;
  dst += y*surface->pitch;
  int* ptr = (int*) dst;
  ptr[x] = c;
  }

void TDisplay::draw() {
  SDL_LockSurface(surface);

  for(int y=0; y<g.sizeY; y++)
  for(int x=0; x<g.sizeX; x++) if(g.map[y][x] != lm[y][x] || g.owner[y][x] != lo[y][x] || g.map[y][x] == cExplosion) {
    int tx = x * Res;
    int ty = y * Res;
    CellType m = g.map[y][x];
    if(m == cBig) {
      for(int ky=0; ky<2; ky++) for(int kx=0; kx<2; kx++) {
        m = g.map[y-ky][x-kx];
        if(isBonus(m) || isCannon(m)) {
          tx -= Res * kx;
          ty -= Res * ky;
          goto bre;
          }
        }
      bre: ;
      }
    TPlayer* pl = g.pl[g.owner[y][x]];
    IconId id = 
      m == cFree ? (g.owner[y][x] == 0 ? icFree : icIncastle) :
      m == cBonusLevel ? icLevel :
      m == cBonusAmmo ? icAmmo :
      m == cBonusScore ? icScore :
      m == cCannonF ? icCannon :
      m == cCannonA ? icCannonA :
      m == cPermaWall ? icPerma :
      m == cBomb1 ? icBomb1 :
      m == cBomb2 ? icBomb2 :
      m == cExplosion ? IconId(int(icExplosion) + ((g.time / 125) & 3)) :
      m == cXunux ? icXunux :
      icBreak;

    pl->icons[int(id)].drawAt(tx, ty);

    lo[y][x] = g.owner[y][x];
    lm[y][x] = g.map[y][x];
    }
  
  for(int o=0; o<size(g.objects); o++)
    if(g.objects[o]) g.objects[o]->draw();

  SDL_UnlockSurface(surface);

  int y0 = sizeY - 20 * size(g.pl) - 7;
  int step = (sizeX-20)/7;
  putText(10, y0, 0xFFFFFF, "TYPE", 0);
  putText(10+3*step, y0, 0xFFFFFF, "LIVES", 2);
  putText(10+4*step, y0, 0xFFFFFF, "FRAGS", 2);
  putText(10+5*step, y0, 0xFFFFFF, "SCORE", 2);
  putText(10+6*step, y0, 0xFFFFFF, "LEVEL", 2);
  putText(10+7*step, y0, 0xFFFFFF, "AMMO", 2);
  for(int p=1; p<size(g.pl); p++) {
    TPlayer *pla = g.pl[p];
    int c = pla->color;
    int y = y0 + 20*p;

    for(int u=2; u<8; u++) {
      putText(10+u*step, y, 0, "||||||||||||||", 2);
      putText(10+u*step+2, y, 0, "||||||||||||||", 2);
      }
    putText(10, y, 0, "|||||||||||||||||||||||||||||||", 0);
    putText( 8, y, 0, "|||||||||||||||||||||||||||||||", 0);

    putText(10, y, c, pla->getTypeName(), 0);
    if(pla->typeval != -1) putText(10+2*step, y, c, qitoa(pla->typeval), 2);
    putText(10+3*step, y, c, qitoa(pla->lives), 2);
    putText(10+4*step, y, c, qitoa(pla->frags), 2);
    putText(10+5*step, y, c, qitoa(pla->score), 2);
    putText(10+6*step, y, c, qitoa(pla->level), 2);
    putText(10+7*step, y, c, qitoa(pla->ammo), 2);
    }

  SDL_UpdateRect(surface, 0, 0, 0, 0);
  }

void TDisplay::putText(int x, int y, int color, const char *text, int align) {
  if(text[0] == 0) return;
  if(d.font) {
    // Some versions of SDL_ttf don't accept spaces in text.
    // This is an ugly hack to fix it.
    char text2[640];
    strncpy(text2, text, 640);
    for(int k=0; k<640; k++) if(text2[k] == ' ') text2[k] = '_';
    
    SDL_Color col;
    col.r = 0;
    col.g = 0;
    col.b = 1;
    SDL_Surface *txt = TTF_RenderText_Solid(d.font, text, col);
    if(!txt) txt = TTF_RenderText_Solid(d.font, text2, col);
    if(!txt) return;
    SDL_Rect rect;
    rect.y = y;
    rect.h = sizeY - y;
    rect.w = txt->w;
    switch(align) {
      case 0: rect.x = x; break;
      case 1: rect.x = x - rect.w/2; break;
      case 2: rect.x = x - rect.w; break;
      }
      
    rect.x--; SDL_BlitSurface(txt, NULL,d.surface,&rect); rect.x++;
    rect.x++; SDL_BlitSurface(txt, NULL,d.surface,&rect); rect.x--;
    rect.y--; SDL_BlitSurface(txt, NULL,d.surface,&rect); rect.y++;
    rect.y++; SDL_BlitSurface(txt, NULL,d.surface,&rect); rect.y--;
    SDL_FreeSurface(txt);
    
    col.r = color >> 16;
    col.g = color >> 8;
    col.b = color;
    txt = TTF_RenderText_Solid(d.font, text, col);
    if(!txt) txt = TTF_RenderText_Solid(d.font, text2, col);
    if(!txt) return;
    SDL_BlitSurface(txt, NULL,d.surface,&rect);
    SDL_FreeSurface(txt);
    }
  }

void TDisplay::menuLoop() {
  while(!stop) {
    SDL_Event event;
    bool ev = SDL_WaitEvent(&event);
    if(ev) switch (event.type) {

      case SDL_KEYDOWN:
        menuKey(event.key.keysym.sym);
        menuDraw();
        break;
      
      case SDL_MOUSEMOTION: {
        int mp = event.motion.y / 20 - 4;
        if(mp >= 0 && mp <= 14 + size(lps))
          menupos = mp;
        menuDraw();
        break;
        }

      case SDL_MOUSEBUTTONDOWN:
        menuKey(event.button.button == 1 ? KEY_MOUSELEFT : KEY_MOUSERIGHT);
        menuDraw();
        break;

      case SDL_QUIT:
        stop = true;
        return;
      
      default:
        menuDraw();
      }
    }
  }

int goodPos(int i) { return d.menupos == i ? 0xFFFFFF : 0x0000FF; }

char *gameTypeName(TGameType t) {
  if(t == gLocal) return "LOCAL";
  if(t == gHost) return "SERVER";
  if(t == gClient) return "CLIENT";
  return "UNKNOWN";
  }

void drawMenuLine(int y,const char *text, const char *option) {
  int r = d.sizeX / 2;
  d.putText(r-20, 80+20*y, goodPos(y), text, 2);
  d.putText(r+20, 80+20*y, goodPos(y), option, 0);
  }

void drawMenuLine(int y,const char *text, int val) {
  char buf[16];
  sprintf(buf, "%d", val);
  drawMenuLine(y, text, buf);
  }

void TDisplay::menuDraw() {
  SDL_LockSurface(surface);

  for(int y=0; y<sizeY; y++)
  for(int x=0; x<sizeX; x++) putpix(x, y, (((x ^ y) & 0x3F) << 8));

  SDL_UnlockSurface(surface);

  int r = sizeX / 2;
  putText(r, 20, 0x00FF00, "UNTAHRIS 3.1", 1);
  putText(r, 40, 0x00FF00, "Common Playground", 1);
  
  drawMenuLine(0, "Multiplayer [L/S/C]", gameTypeName(c.gt));
  drawMenuLine(1, "Server", c.server.c_str());
  drawMenuLine(2, "port", c.port);
  drawMenuLine(3, "screen resolution X", c.resX);
  drawMenuLine(4, "screen resolution Y", c.resY);
  drawMenuLine(5, "full screen", c.fs ? "YES" : "NO");
  drawMenuLine(6, "cell size in pixels", c.res);
  drawMenuLine(7, "width %", c.pctX);
  drawMenuLine(8, "height %", c.pctY);
  drawMenuLine(9, "lines for scores", c.scorelines);
  drawMenuLine(10, "initial world", worldNames[c.initialWorld]);
  
  if(c.initialWorld != wtLoad)
    drawMenuLine(11, "bonuses at start", c.initBonus);
  else
    drawMenuLine(11, "map to load", c.mapname.c_str());

  drawMenuLine(12, "graphics style", gtNames[c.graphicsType]);
  drawMenuLine(13, "background", bgNames[c.background]);

  for(int p=0; p<size(lps); p++) {
    bool onP = menupos == 14+p;
    char buf[80];
    if(!onP) sprintf(buf, "edit a player");
    else sprintf(buf, "[C]hange color [RGB], redefine [K]eys, [D]elete");
    putText(r, 360 + 20*p, lps[p].color, buf, 1);
    }
  putText(r, 360 + 20*size(lps), goodPos(14+size(lps)), "create a new player", 1);
  
  putText(r, d.sizeY-40, 0xFF0000, c.err.c_str(), 1);

  SDL_UpdateRect(surface, 0, 0, 0, 0);
  }

void checkConfig() {
  c.err = "press ENTER to start the game!";
  c.ok = true;

  if(c.res < 1)
    c.ok = false, c.err = "Bad cell resolution";
  else if(c.res > 32)
    c.ok = false, c.err = "Bad cell resolution (32 is max)";
  }

void affectString(int key, string& s) {
  if(key > 128 || key < 0) return;
  if(key == 8) {
    if(s.size() <= 1) s = "";
    else s.erase(s.size() - 1);
    }
  else s += char(key);
  }

template<class T> void affectToggle(int key, T& x, int vals) {
  char* ptr = (char*) (&x);
  if(key & 1) (*ptr)++;
  else (*ptr) += (vals-1);
  (*ptr) %= vals;
  }

void affectGameType(int key, TGameType& gt) {
  if(key == 'l') gt = gLocal;
  if(key == 'h') gt = gHost;
  if(key == 's') gt = gHost;
  if(key == 'c') gt = gClient;
  if(key == 'j') gt = gClient;
  }

void affectInt(int key, int& val) {
  if(key >= '0' && key <= '9') {
    val *= 10;
    val += key - '0';
    }
  else if(key == 8) val /= 10;
  else if(key == '+' || key == SDLK_KP6 || key == SDLK_RIGHT || key == KEY_MOUSERIGHT)
    val++;
  else if(key == '-' || key == SDLK_KP4 || key == SDLK_LEFT || key == KEY_MOUSELEFT)
    val--;
  }

void editKeysFor(TLocalPlayer& lp) {
  char *keyNames[8] = {
    "UP",
    "RIGHT",
    "DOWN",
    "LEFT",
    "FIRE",
    "EARTH",
    "AIR",
    "WATER"
    };
  for(int k=0; k<8; k++) {
    SDL_LockSurface(d.surface);
  
    for(int y=0; y<d.sizeY; y++)
    for(int x=0; x<d.sizeX; x++) d.putpix(x, y, (((x+y)^(x-y))&0xFF) << 16);
  
    SDL_UnlockSurface(d.surface);
  
    d.putText(d.sizeX / 2, d.sizeY / 2 - 10, 0x00FFFF, keyNames[k], 1);

    SDL_UpdateRect(d.surface, 0, 0, 0, 0);

    while(true) {
      SDL_Event event;
      bool ev = SDL_PollEvent(&event);
      if(ev) switch (event.type) {
  
        case SDL_KEYDOWN:
          lp.keys[k] = event.key.keysym.sym;
          if(lp.keys[k] == 27) return;
          goto nextKey;
  
        case SDL_QUIT:
          return;
        }
      }
    nextKey: ;
    }
  }

void affectPlayer(int key, int index) {
  switch(key) {
    case 'c':
      lps[index].color = randColor();
      break;
    case 'r':
      lps[index].color ^= (rand() & 0xFF) << 16;
      break;
    case 'g':
      lps[index].color ^= (rand() & 0xFF) << 8;
      break;
    case 'b':
      lps[index].color ^= (rand() & 0xFF);
      break;
    case 'k':
      editKeysFor(lps[index]);
      break;
    case 8:
    case 'd':
    case 'e':
      lps.erase(lps.begin()+index, lps.begin()+index+1);
      break;
    }
  }

void TDisplay::menuKey(int key) {
  switch(key) {
    case 13: {
      if(needRestart) setMode(), needRestart = false;
      checkConfig();
      if(c.ok)
        startGame();
      stop = false;
      break;
      }
    case 27:
      stop = true;
      break;
    case SDLK_KP2: case SDLK_UP:
      menupos--;
      if(menupos < 0) menupos = 0;
      break;
    case SDLK_KP8: case SDLK_DOWN:
      menupos++;
      if(menupos > 14 + size(lps)) menupos = 14 + size(lps);
      break;
    default:
      if(menupos==0) affectGameType(key, c.gt);
      if(menupos==1) affectString(key, c.server);
      if(menupos==2) affectInt(key, c.port);

      if((menupos == 3 || menupos == 4) && key == ' ') {
        if(c.resX == 800) c.resX = 1024, c.resY = 768;
        else c.resX = 800, c.resY = 600;
        }
      if(menupos==3) affectInt(key, c.resX);
      if(menupos==4) affectInt(key, c.resY);
      if(menupos==5) affectToggle(key, c.fs, 2);
      if(menupos==6) affectInt(key, c.res);
      if(menupos==7) affectInt(key, c.pctX);
      if(menupos==8) affectInt(key, c.pctY);
      if(menupos==9) affectInt(key, c.scorelines);
      if(menupos==10) affectToggle(key, c.initialWorld, numWorldTypes);
      if(menupos==11 && c.initialWorld != wtLoad) affectInt(key, c.initBonus);
      if(menupos==11 && c.initialWorld == wtLoad) affectString(key, c.mapname);
      if(menupos==12) affectToggle(key, c.graphicsType, 3);
      if(menupos==13) affectToggle(key, c.background, 5);
      if(menupos>= 14 && menupos < 14 + size(lps))
        affectPlayer(key, menupos - 14);
      else if(menupos==14+size(lps)) {
        TLocalPlayer newOne;
        newOne.color = randColor();
        editKeysFor(newOne);
        lps.push_back(newOne);
        }
      if(menupos == 3 || menupos == 4 || menupos == 8)
        needRestart = true;
      checkConfig();
    }
  }

int TDisplay::getCurrentTime() {
  return SDL_GetTicks() + delta;
  }

void TDisplay::setCurrentTime(int val) {
  delta = val - SDL_GetTicks();
  }

#define dlCircle 1
#define dlPlus   2
#define dlCross  4

void TDisplay::initForGame() {
  
  d.Res = c.res;

  int dr1 = (d.Res-1)/2, dr2 = dr1+1, dr3 = d.Res-1;

  // background color
  int bak = -2; if(c.background == bgGray) bak = 0x404040;

  bool line = c.graphicsType == gtLines;

  int wid = c.graphicsType == gtGradient ? 1 : 2;
  
  Ic(icIncastle).fillWith("Z", d.Res);
  if(c.graphicsType == gtGradient) {
    for(int x=0; x<d.Res; x++) for(int y=0; y<d.Res; y++) Ic(icIncastle).i.pixel[y][x] = 
      cmix(0, 0x80000000, (x+y-d.Res+1) / (d.Res * 1.1));
    }
  if(c.graphicsType == gtRetro) {
    Ic(icIncastle).fillWith("pu", d.Res);
    }
  if(c.graphicsType == gtLines) {
    Ic(icIncastle).gradFill(dr1, dr1, dr2, dr2, 0xFF000000, 0xFF000000);
    }
  
  Ic(icFree).fillWith("Z", d.Res);
  if(c.background == bgGray) Ic(icFree).fillWith("k", d.Res), Ic(icIncastle) = Ic(icFree);
  if(c.background == bgGrid) Ic(icFree) = Ic(icIncastle);
  
  Ic(icPerma).fillWith(line ? "P" : "W", d.Res);
  
//Ic(icPerma).gradFill(1, 1, dr3-1, dr3-1, 0x80000000, 0xFFFFFF);
  int c1 = 0x80000000, c2 = 0xFFFFFF;
  if(c.graphicsType == gtLines) c1 = c2 = bak;
  Ic(icPerma).gradFill(wid, wid, dr1-wid, dr1-wid, c1, c2);
  Ic(icPerma).gradFill(wid, dr2+wid, dr1-wid, dr3-wid, c1, c2);
  Ic(icPerma).gradFill(dr2+wid, wid, dr3-wid, dr1-wid, c1, c2);
  Ic(icPerma).gradFill(dr2+wid, dr2+wid, dr3-wid, dr3-wid, c1, c2);
  
/*Ic(icFree).fillWith("d", d.Res);
  Ic(icFree).gradFill(0, 0, dr3, dr3, 0x00000000, 0x80000000);
  Ic(icFree).gradFill(1, 1, dr3-1, dr3-1, 0x40000000, 0x80000000); */

  if(c.graphicsType == gtLines) {
    Ic(icBreak).fillWith("P", d.Res);
    Ic(icBreak).gradFill(2, 2, dr3-2, dr3-2, bak, bak);
    }
  else {
    Ic(icBreak).fillWith("p", d.Res);
    Ic(icBreak).gradFill(wid, wid, dr3-wid, dr3-wid, 0xFF000000, 0x80000000);
    }

  Ic(icBall).fillWith("0", d.Res);
  Ic(icBall).gradBall(0, 0, dr3-1, dr3-1, 0, 0x80FFFFFF);
  if(c.graphicsType == gtLines) {
    Ic(icBall).gradBall(0, 0, dr3-1, dr3-1, 0x80FFFFFF, 0x80FFFFFF);
    Ic(icBall).gradBall(2, 2, dr3-3, dr3-3, -1, -1);
    }

  Ic(icBullet).fillWith("0", 8);
  Ic(icBullet).gradBall(0, 0, 7, 7, 0, 0x80FFFFFF);
  if(c.graphicsType == gtLines) {
    Ic(icBullet).gradBall(0, 0, 7, 7, 0x80FFFFFF, 0x80FFFFFF);
    Ic(icBall).gradBall(2, 2, dr3-3, dr3-3, -1, -1);
    }

  int lft0 = d.Res < 24 ? wid : 2;
  int lft1 = 2*lft0;
  int lft2 = 3*lft0;
  int rgt0 = d.Res*2-1-lft0;
  int rgt1 = d.Res*2-1-lft1;
  int rgt2 = d.Res*2-1-lft2;
  
  for(int t=icAmmo; t<=icScore; t++) {
    int col = 
      t == icAmmo ? 0xFF0000 : t == icScore ? 0xFF00 : 0xFF;\

    Ic(t).fillTile(Ic(icFree), d.Res*2);
    Ic(t).gradBall(lft0, lft0, rgt0, rgt0, 0x00000000, 0x00000000);
    Ic(t).gradBall(lft1, lft1, rgt1, rgt1, 0xFF000000, 0xFF000000);
    Ic(t).gradBall(lft2, lft2, rgt2, rgt2, 0x00000000, col);

    if(c.graphicsType == gtLines) {
      Ic(t).gradBall(lft0, lft0, rgt0, rgt0, col, col);
      Ic(t).gradBall(lft1, lft1, rgt1, rgt1, bak, bak);
      Ic(t).gradFill(d.Res-1, d.Res-1, d.Res, d.Res, 0xFF000000, 0xFF000000);
      }
    }
  
  Ic(icCastle).fillWith("pY0P0P0", d.Res);
  Ic(icCannon).fillWith("dhhYPYPYP", d.Res*2);
  Ic(icCannonA).fillWith("dhhRPRPRP", d.Res*2);
  Ic(icCannonP).fillWith("000RPRPRP", d.Res*2);

  if(line) {
    Ic(icSpawn).fillWith("PP0", d.Res+3);
    }
  else {
    Ic(icSpawn).fillWith("pPp0", d.Res+3);
    Ic(icSpawn).gradFill(0, 0, dr3+3, dr3+3, 0, 0x80000000);
    Ic(icSpawn).gradFill(1, 1, dr3+2, dr3+2, 0x80000000, 0xFF000000);
    Ic(icSpawn).gradFill(2, 2, dr3+1, dr3+1, 0, 0x80000000);
    Ic(icSpawn).gradFill(3, 3, dr3+0, dr3+0, -1, -1);
    }

  Ic(icTarget).fillWith("0", d.Res | 1);
  int tx = (d.Res+1) / 2;
  Ic(icTarget).gradFill(tx-1, 0, tx+1, tx*2, 0, 0);
  Ic(icTarget).gradFill(0, tx-1, tx*2, tx+1, 0, 0);
  Ic(icTarget).gradFill(tx, 0, tx, tx*2, 0x80000000, 0xFF000000);
  Ic(icTarget).gradFill(0, tx, tx*2, tx, 0x80FFFFFF, 0xFF000000);

  Ic(icBomb1).fillTile(Ic(icFree), d.Res);
  Ic(icBomb1).gradBall(0, 0, dr3, dr3, 0xFF0000, 0xFF0000);
  Ic(icBomb1).gradBall(1, 1, dr3-1, dr3-1, 0, 0x80000000);

  Ic(icBomb2).fillTile(Ic(icFree), d.Res);
//Ic(icBomb2).fillWith("dhphY", d.Res);
  Ic(icBomb2).gradBall(0, 0, dr3, dr3, 0xFFFF00, 0xFFFF00);
  Ic(icBomb2).gradBall(1, 1, dr3-1, dr3-1, 0, 0x80000000);

  if(c.graphicsType == gtLines) {
    Ic(icBomb1).fillWith("Z", d.Res);
    Ic(icBomb2).fillWith("Z", d.Res);
    Ic(icBomb1).drawLines(0,0,dr3,dr3, 0x80800000,dlCross);
    Ic(icBomb2).drawLines(0,0,dr3,dr3, 0x80808000,dlCross);
    }
  
  int bombr = d.Res*3/4;
  Ic(icBomber).fillWith("0", bombr);

  if(c.graphicsType == gtLines) {
    Ic(icBomber).drawLines(0, 0, bombr-1, bombr-1, 0xFF000000, dlCircle);
    }
  else {
    Ic(icBomber).gradBall(0, 0, bombr-1, bombr-1, 0, 0);
    Ic(icBomber).gradBall(1, 1, bombr-2, bombr-2, 0x80000000, 0xFF000000);
    }

  if(c.graphicsType == gtLines) {
    for(int t=icExplosion; t<=icExplosion3; t++) {
      Ic(t).fillWith("Z", d.Res);
      }
    Ic(icExplosion).drawLines(0,0,dr3,dr3,0x80FF0000,dlCross | dlPlus);
    Ic(icExplosion1).drawLines(0,0,dr3,dr3,0x60FF8000,dlCross | dlPlus);
    Ic(icExplosion2).drawLines(0,0,dr3,dr3,0x40FFFF00,dlCross | dlPlus);
    Ic(icExplosion3).drawLines(0,0,dr3,dr3,0x60FF8000,dlCross | dlPlus);
    }
  else {
    Ic(icExplosion).fillWith("R", d.Res);
    Ic(icExplosion).gradFill(0,0,dr3,dr3,0x80FF0000,0x80FFFF00);
    Ic(icExplosion1).fillWith("R", d.Res);
    Ic(icExplosion1).gradFill(0,0,dr3,dr3,0x60FF8000,0x60FF8000);
    Ic(icExplosion2).fillWith("R", d.Res);
    Ic(icExplosion2).gradFill(0,0,dr3,dr3,0x40FFFF00,0x40FF0000);
    Ic(icExplosion3).fillWith("R", d.Res);
    Ic(icExplosion3).gradFill(0,0,dr3,dr3,0x60FF8000,0x60FF8000);
    }

  if(c.graphicsType == gtLines) {
    Ic(icXunux).fillWith("Z", d.Res);
    Ic(icXunux).drawLines(0, 0, dr3, dr3, 0xFF000000, dlPlus);
    }
  else {
    Ic(icXunux).fillWith("R", d.Res);
    Ic(icXunux).gradFill(0, 0, dr3, dr3, 0xFF000000, 0x00FFFFFF);
    }

  Ic(icThrust).fillWith("0", d.Res*2);
  for(int y=0; y<d.Res*3/2; y++) {
    int w = (y * 14) / 30;
    int x0 = d.Res-1-w;
    int x1 = d.Res+w;
    for(int x=x0; x <= x1; x++) {
      bool boundary = x <= x0+1 || x>=x1-1 || y >= d.Res*3/2-1;
      
      int col;
      
      if(c.graphicsType == gtRetro)
        col = boundary ? 0 : 0xFF000000;

      if(c.graphicsType == gtLines)
        col = boundary ? int(0xFF000000) : -1;

      if(c.graphicsType == gtGradient)      
        col = cmix(0xFF000000, 0, 
        ((x-x0)/(x1-x0+.0) * (x-x1)/(x0-x1+.0)) * 4
        );
      
      Ic(icThrust).i.pixel[y][x] = col;
      }
    }
    
  Ic(icWorm).fillWith("0", d.Res);
  Ic(icWorm).gradFill(d.Res*3/8, 0, d.Res*5/8, d.Res-1, 0xFF000000, 0x00000000);

  Ic(icBazooka).fillWith("0", d.Res);
  Ic(icBazooka).gradFill(d.Res*4/8-1, 0, d.Res*4/8, d.Res*3/4, 0x80FFFFFF, 0xFFFFFFFF);

  char buf[64];
  sprintf(buf, "graph%d%s.unt", d.Res, gtNames[c.graphicsType]);
  FILE *f = fopen(buf, "rt");
  if(f) {
    while(1) {
      string s = readStr(f);
      if(s == "") break;
      bool found = false;
      for(int k=0; k<numIcons; k++)
        if(s == iconnames[k]) found = true, icons[k].loadFrom(f);
      if(!found && s[0] != '#') printf("Unknown icon name '%s'\n", s.c_str());
      }
    fclose(f);
    }
  }

void TDisplay::initBackground() {
  d.ofsX = (d.sizeX - c.res * g.sizeX) / 2;
  d.ofsY = (d.sizeY - 20 * c.scorelines - c.res * g.sizeY) / 2;
  
  int sy = c.res * g.sizeY, sx = c.res * g.sizeX;
  if(bground) delete bground;

  bground = new int[sx*sy];
  switch(c.background) {
    case bgPlasma:
      bground[0] = randDarkColor();
      bground[sx-1] = randDarkColor();
      bground[sy*sx-1] = randDarkColor();
      bground[sy*sx-sx] = randDarkColor();
      genBGround(0, 0, sx-1, sy-1, sx);
      return;
    
    case bgBlack: case bgGrid:
      for(int x=0; x<sx*sy; x++) bground[x] = 0;
      break;
    
    case bgGray:
      for(int x=0; x<sx*sy; x++) bground[x] = 0x404040;
      break;
    
    case bgGradient:
      for(int y=0; y<sy; y++)
      for(int x=0; x<sx; x++) {
        int r = int((64 * y)/sy);
        // int r = int(64 * (1+sin(y/20.0+x/30.0)));
        // int g = int(64 * (1+sin(y/50.0-x/10.0)));
        // int b = int(64 * (1+sin(y/20.0-x/40.0)));
        bground[y*sx+x] = (r << 16); // + (g << 8) + (b);
        }
      break;
    }
  }

void TDisplay::genBGround(int x0, int y0, int x1, int y1, int sx) {
  if(x1 == x0+1 && y1 == y0+1) return;
  if(x1-x0 > y1-y0) {
    int x2 = (x0+x1)/2;
    bground[x2+sx*y0] = cmix(randDarkColor(), cmix(bground[x0+sx*y0], bground[x1+sx*y0], .5), (20 / (20 + sqrt(x1-x0))));
    bground[x2+sx*y1] = cmix(randDarkColor(), cmix(bground[x0+sx*y1], bground[x1+sx*y1], .5), (20 / (20 + sqrt(x1-x0))));
    genBGround(x0,y0,x2,y1,sx);
    genBGround(x2,y0,x1,y1,sx);
    }
  else {
    int y2 = (y0+y1)/2;
    bground[x0+sx*y2] = cmix(randDarkColor(), cmix(bground[x0+sx*y0], bground[x0+sx*y1], .5), (20 / (20 + sqrt(y1-y0))));
    bground[x1+sx*y2] = cmix(randDarkColor(), cmix(bground[x1+sx*y0], bground[x1+sx*y1], .5), (20 / (20 + sqrt(y1-y0))));
    genBGround(x0,y0,x1,y2,sx);
    genBGround(x0,y2,x1,y1,sx);
    }
  }

int TDisplay::interpretPixel(char ch) {
  switch(ch) {
    case 'd': return 0;
    case 'h': return 0x000040;
    case 'W': return 0xFFFFFF;
    case 'w': return 0xC0C0C0;
    case 'K': return 0x808080;
    case 'k': return 0x404040;
    case 'P': return 0xFF000000;
    case 'p': return 0x80000000;
    case 'u': return 0x40000000;
    case 'R': return 0xFF0000;
    case 'r': return 0x800000;
    case 'G': return 0x00FF00;
    case 'g': return 0x008000;
    case 'B': return 0x0000FF;
    case 'b': return 0x000080;
    case 'Y': return 0xFFFF00;
    case 'y': return 0x808000;
    case 'O': return 0xFF8000;
    case 'o': return 0x804000;
    case '0': return -1;
    case 'Z': return -2;
    }
  }

void TDisplayIcon::fillWith(const char *chars, int _size) {
  int len = strlen(chars);
  i.size = _size;
  for(int x=0; x<i.size; x++)
  for(int y=0; y<i.size; y++) {
    int k = x;
    if(y < k) k = y;
    if(i.size-y-1 < k) k = i.size-y-1;
    if(i.size-x-1 < k) k = i.size-x-1;
    if(k >= len) k = len-1;
    i.pixel[y][x] = d.interpretPixel(chars[k]);
    }
  }

void TDisplayIcon::fillTile(TDisplayIcon& ic, int _size) {
  i.size = _size;
  for(int x=0; x<i.size; x++)
  for(int y=0; y<i.size; y++) {
    i.pixel[y][x] = ic.i.pixel[y%ic.i.size][x%ic.i.size];
    }
  }

void TDisplayIcon::gradFill(int x0, int y0, int x1, int y1, int c0, int c1) {
  if(c.graphicsType == gtRetro) c1 = c0;
  for(int x=x0; x<=x1; x++)
  for(int y=y0; y<=y1; y++)
    i.pixel[y][x] = cmix(c0, c1, (x+y-x0-y0) / (x1+y1-x0-y0-.0));
  }

void TDisplayIcon::gradBall(int x0, int y0, int x1, int y1, int c0, int c1) {
  if(c.graphicsType == gtRetro) c0 = c1;
  for(int x=x0; x<=x1; x++)
  for(int y=y0; y<=y1; y++) {
    float fx = (x+x - x0-x1) / (x1-x0+.0);
    float fy = (y+y - y0-y1) / (y1-y0+.0);
    float f = fx*fx+fy*fy;
    if(f <= 1)
      i.pixel[y][x] = cmix(c1, c0, f);
    }
  }

void TDisplayIcon::drawLines(int x0, int y0, int x1, int y1, int color, int type) {
  float minf = (2+2-x0-x1) / (x1-x0+.0); minf *= minf;

  for(int x=x0; x<=x1; x++)
  for(int y=y0; y<=y1; y++) {
    bool on = false;
    float fx = (x+x - x0-x1) / (x1-x0+.0);
    float fy = (y+y - y0-y1) / (y1-y0+.0);
    float f = fx*fx+fy*fy;
    
    if((type & dlCircle) && f <= 1 && f >= minf) on = true;
    if((type & dlCross) && f <= 1 && 
      ((x >= y-y0+x0-1 && x <= y-y0+x0+1) || (x >= x0+y1-y-1 && x <= x0+y1-y+1)))
        on = true;
    if((type & dlPlus) && ((x >= (x0+x1)/2 && x <= (x0+x1+1)/2) || (y >= (y0+y1)/2 && y <= (y0+y1+1)/2)))
      on = true;
        
    if(on)
      i.pixel[y][x] = color;
    }
  }

void TDisplayIcon::loadFrom(FILE *f) {
  string line = readStr(f);
  if(line.size() > 64) { printf("Image too big\n"); return; }
  i.size = line.size();
  for(int y=0; y<i.size; y++) {
    if(y) line = readStr(f);
    if(line.size() > 64) { printf("Line too long\n"); return; }
    if(int(line.size()) != i.size) {printf("Misformed line\n"); return; }
    for(int x=0; x<int(line.size()); x++)
      i.pixel[y][x] = d.interpretPixel(line[x]);
    }
  }

void TPlayerIcon::drawAt(int ox, int oy) {
  
  for(int x=0; x<i.size; x++)
  for(int y=0; y<i.size; y++) {
    int ch = i.pixel[y][x];
    if(ch == -2)
      d.putpix(d.ofsX+ox+x, d.ofsY+oy+y, d.bground[(oy+y)*c.res*g.sizeX+ox+x]);
    else if(ch != -1)
      d.putpix(d.ofsX+ox+x, d.ofsY+oy+y, ch);
    }
  }

void TPlayerIcon::drawAtFlipped(int ox, int oy) {
  int boffs = d.ofsX + c.res*g.sizeX*d.ofsY;

  for(int x=0; x<i.size; x++)
  for(int y=0; y<i.size; y++) {
    int ch = i.pixel[y][x];
    if(ch == -2)
      d.putpix(d.ofsX+ox+i.size-x-1, d.ofsY+oy+y, d.bground[(oy+y)*c.res*g.sizeX+ox+x]);
    else if(ch != -1)
      d.putpix(d.ofsX+ox+i.size-x-1, d.ofsY+oy+y, ch);
    }
  }

void TPlayerIcon::assignColor(int c) {
  for(int x=0; x<i.size; x++)
  for(int y=0; y<i.size; y++) {
    int col = i.pixel[y][x];
    if(col != -1 && col != -2)
      i.pixel[y][x] = cmix(col, c, ((col >> 24) & 0xFF) / 255.0);
    }
  }

void TPlayerIcon::drawAtRotated(int ox, int oy, int rx, int ry) {
  for(int x=0; x<i.size; x++)
  for(int y=0; y<i.size; y++) {
    int ch = i.pixel[y][x];
    if(ch != -1) {
      int x0 = x*2-i.size;
      int y0 =  y*2-i.size;
      if(i.size&1 == 0) x0 += 1, y0 += 1;
      d.putpix(d.ofsX+ox-((x0*ry+y0*rx+127)>>8), d.ofsY+oy-((y0*ry-x0*rx+127)>>8), ch);
      }
    }
  }
