// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// TPlayer and TPlayObject implementation.

struct OSpawner : TPlayObject {
  public:
  int16 spx, spy; // spawn location

  OSpawner(int _owner) {
    owner = _owner;
    time = g.time;
    spx = 1 + (g.rand() % (g.sizeX - 2));
    spy = 1 + (g.rand() % (g.sizeY - 2));
    g.objects.push_back(this);
    }

  OSpawner(TBuffer &b) {
    b.get(owner); b.get(spx); b.get(spy); b.get(time);
    }
  void save(TBuffer &b) {
    b.putChar(14);
    b.put(owner); b.put(spx); b.put(spy); b.put(time);
    }

  void newCharacter() {
    // level is reset for each new addition
    d.playSoundAt(d.sndBirth, spx, spy);
    ow()->lives++;
    ow()->level = 0;
    ow()->ammo = ec().initAmmo;
    ow()->typeval = 1;  
    time = g.time + 1000000;
    }
  
  void spawn() {
    int32& typeval (ow()->typeval);
    int8 index = ow()->index;
    
    if(typeval == -1 && !ow()->editing) switch(ow()->charid) {
      case 0:
        if(g.map[spy][spx] != cFree) return;
        d.playSoundAt(d.sndBirth, spx, spy);
        new OSnakeHead(index, spx, spy, dirAt(spx, spy), true);
        newCharacter();
        return;
 
      case 1:
        if(g.map[spy][spx] != cFree) return;
        new OSnakeHead(index, spx, spy, dirAt(spx, spy), false);
        newCharacter();
        return;
 
      case 2:
        new OTetris(index, spx-1, spy);
        newCharacter();
        return;
 
      case 3:
        if(spy < 2 || spy > g.sizeY - 2) return;
        if(spx < 2 || spx > g.sizeX - 2) return;
        new OThruster(index, spx, spy);
        newCharacter();
        typeval = gc.initFuel;
        return;
      
      case 4:
        new OBomber(index, Ctr(spx), Ctr(spy));
        newCharacter();
        typeval = gc.bombInitRange;
        return;
      
      case 5:
        if(g.map[spy][spx] != cFree) return;
        g.mapSet(spx, spy, cCastleWall, index);
        new OCastle(index, spx, spy);
        newCharacter();
        typeval = gc.castleInitEnergy;
        return;
        
      case 6:
        new OWorm(index, spx, spy);
        newCharacter();
        return;
 
      case 7:
        if(spy < 4 || spy > g.sizeY - 4) return;
        if(spx < 4 || spx > g.sizeX - 4) return;
        new OPad(index, Ctr(spx), Ctr(spy));
        newCharacter();
        return;
 
      case 8:
        new OXunux(index, spx, spy, dirAt(spx, spy));
        newCharacter();
        return;
 
      case 9:
        d.playSoundAt(d.sndPlace, spx, spy);
        new OBall(index, Ctr(spx), Ctr(spy), g.rand() % 360);
        return;
      }
  
    if(ow()->editing) switch(ow()->editid) {
 
      case 0:
        d.playSoundAt(d.sndPlace, spx, spy);
        new OBall(0, Ctr(spx), Ctr(spy), g.rand() % 360);
        return;
 
      case 1:
        d.playSoundAt(d.sndPlace, spx, spy);
        if(mapFreeAt(spx, spy)) g.mapSet(spx, spy, cBreakWall, index);
        return;
 
      case 2:
        d.playSoundAt(d.sndPlace, spx, spy);
        if(mapFreeAt(spx, spy)) g.mapSet(spx, spy, cPermaWall, index);
        return;
 
      case 3:
        d.playSoundAt(d.sndPlace, spx, spy);
        if(mapFreeAt(spx, spy)) g.mapSet(spx, spy, cFree, index);
        return;
      
      case 4:
        d.playSoundAt(d.sndPlace, spx, spy);
        if(mapFreeBigAt(spx, spy)) placeBig(spx, spy, cCannonF, index);
        return;
        
      case 5:
        d.playSoundAt(d.sndRed, spx, spy);
        if(mapFreeBigAt(spx, spy)) placeBig(spx, spy, cBonusAmmo, index);
        return;
             
      case 6:
        d.playSoundAt(d.sndBlue, spx, spy);
        if(mapFreeBigAt(spx, spy)) placeBig(spx, spy, cBonusLevel, index);
        return;
        
      case 7:
        d.playSoundAt(d.sndGreen, spx, spy);
        if(mapFreeBigAt(spx, spy)) placeBig(spx, spy, cBonusScore, index);
        return;
        
      case 8:
        d.playSoundAt(d.sndBirth, spx, spy);
        g.generate();
        return;
      
      case 9:
        d.playSoundAt(d.sndDeath, spx, spy);
        g.destroyBalls();
        return;
      
      case 10:
        d.playSoundAt(d.sndBirth, spx, spy); 
        g.saveMap();
        return;
      
      case 11:
        g.paused = !g.paused;
        return;

      case 12:
        d.stop = true;
        return;
      }
    }

  void repKey(PlayerKey pk) {
    int8& editing (ow()->editing);
    int8& id (editing ? ow()->editid : ow()->charid);
    int8 max = editing ? 13 : 10;
    
    if(pk == keyEarth) {
      id--; if(id==-1) id = max-1;
      }
    if(pk == keyAir) {
      id++; id %= max;
      }
    if(pk == keyWater) {
      editing = !editing;
      }
    if(pk == keyFire) spawn();

    if(pk == keyUp    && spy > 1) spy--;
    if(pk == keyDown  && spy < g.sizeY-2) spy++;
    if(pk == keyLeft  && spx > 1) spx--;
    if(pk == keyRight && spx < g.sizeX-2) spx++;
    }
  
  void pausedGo() { go(); }
  
  void delay(int t) { }

  void go() {
    time = NEVER;
    if(ow()->typeval != -1 && !g.paused) return;

    bool dirp = false;

//  for(int k=0; k<8; k++) printf("%d", ow()->key[k]); printf("\n");

    for(int k=7; k>=0; k--) if(ow()->key[k]) {
      repKey(PlayerKey(k));
      dirp = true;
      }

    if(dirp)
      time = g.time + 20;
    }         
  
  void reactToKey(TKeypress& k) {
    if(ow()->typeval == -1 || g.paused) {
      if(!k.pressed) return;
      repKey(PlayerKey(k.key));
      if(k.key < 8) time = g.time + 400;
      
      if(k.key == keyMoveMouse 
      && k.mx >= 1 && k.my >= 1 && k.mx < g.sizeX-1 && k.my < g.sizeY-1) {
        spx = k.mx; spy = k.my;
        }
      }
    }

  void draw() {
    if(ow()->typeval == -1 || g.paused) {
      int dir = (g.time / 4) % 360;
      ow()->icons[int(icSpawn)].drawAtRotated(
        d.Res * spx + d.Res/2, 
        d.Res * spy + d.Res/2,
        degx[dir], degy[dir]
        );
      }
    }
  };

TPlayObject *TBuffer::getObject() {
  char type;
  get(type);
  switch(type) { 
    case 1:  return new OBullet(*this);
    case 2:  return new OSnakeHead(*this);
    case 3:  return new OThruster(*this);
    case 4:  return new OTetris(*this);
    case 5:  return new OSnakeTail(*this);
    case 6:  return new OCastle(*this);
    case 7:  return new OCannonball(*this);
    case 8:  return new OPad(*this);
    case 9:  return new OBall(*this);
    case 10: return new OBomber(*this);
    case 11: return new OBomb(*this);
    case 12: return new OWorm(*this);
    case 13: return new OXunux(*this);
    case 14: return new OSpawner(*this);
    }
  }

void TPlayObject::killedBy(int player) {
  if(player != owner) g.pl[player]->frags++;
  ow()->typeval = -1;
  g.deleteObject(this);
  }

int TPlayObject::spd() {
  TEntityConfig& e ( ec() );
  
  int l = ow()->level;
  if(l > e.maxSpeed) l = e.maxSpeed;
  l *= e.speedBonus;
  l += e.initSpeed;
  if(l < 1) l = 1;
  
  return 7999 / l + 1;
  }

void TPlayer::createIcons() {
  for(int n=0; n<numIcons; n++) {
    icons[n].i = d.icons[n].i;
    icons[n].assignColor(color);
    }
  }

void TPlayer::init(int c) {
  color = c;
  lives = 0;
  frags = 0;
  score = 0;
  level = 0;
  ammo  = 10;
  typeval=-1;
  editing= 0; charid=0; editid=0;
  for(int k=0; k<8; k++) key[k] = false;
  createIcons();
  }

char *TPlayer::getTypeName() { return editing ? editnames[editid] : charnames[charid]; }

