// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// Main file.

#include "untahris.h"

char *configfile = "config.unt";

#include "tools.cpp"
#include "intface.cpp"
#include "net.cpp"

#undef main

#include "bullet.cpp"
#include "castle.cpp"
#include "thrust.cpp"
#include "ball.cpp"
#include "snake.cpp"
#include "tetris.cpp"
#include "bomber.cpp"
#include "worm.cpp"
#include "xunux.cpp"

#include "player.cpp"
#include "pground.cpp"

void defaultConfig() {
  lps.clear();
  c.server = "localhost";
  c.mapname = "saved";
  c.pctX  = 100;
  c.pctY  = 100;
  c.scorelines = 3;
  c.res   = 16;
  c.resX  = 800;
  c.resY  = 600;
  c.port  = 5790;
  
  c.initialWorld = 0;
  c.initBonus = 2;
  c.background = bgGradient;
  c.graphicsType = gtGradient;
  
  c.fs    = false;
  c.centered = false;
  
  TLocalPlayer lp;
  lp.color = randColor();
  lp.keys[int(keyLeft )] = 'o';
  lp.keys[int(keyRight)] = 'p';
  lp.keys[int(keyUp   )] = 'q';
  lp.keys[int(keyDown )] = 'a';
  lp.keys[int(keyFire )] = ' ';
  lp.keys[int(keyEarth)] = '1';
  lp.keys[int(keyAir  )] = '2';
  lp.keys[int(keyWater)] = '3';
  lps.push_back(lp);
  
  TLocalPlayer lq;
  lq.color = randColor();
  lq.keys[int(keyLeft )] = SDLK_LEFT;
  lq.keys[int(keyRight)] = SDLK_RIGHT;
  lq.keys[int(keyUp   )] = SDLK_UP;
  lq.keys[int(keyDown )] = SDLK_DOWN;
  lq.keys[int(keyFire )] = SDLK_RETURN;
  lq.keys[int(keyEarth)] = SDLK_INSERT;
  lq.keys[int(keyAir  )] = SDLK_HOME;
  lq.keys[int(keyWater)] = SDLK_PAGEUP;
  lps.push_back(lq);
  
  for(int et=0; et<9; et++) {
    gc.ec[et].initSpeed    = 200;
    gc.ec[et].speedBonus   = 10;
    gc.ec[et].maxSpeed     = 999;
    gc.ec[et].initAmmo     = 10;
    gc.ec[et].bonusAmmo    = 10;
    gc.ec[et].ammoPerShoot = 2;
    }
  
  gc.ec[tnSNAKE].ammoPerShoot = 5;
  gc.ec[tnUTRON].ammoPerShoot = 5;
  gc.ec[tnXUNUX].ammoPerShoot = 5;
  gc.ec[tnPAD].ammoPerShoot = 2;
  gc.ec[tnCASTLE].ammoPerShoot = 5;
  gc.ec[tnCASTLE].speedBonus = 100;
  gc.ec[tnCASTLE].bonusAmmo = 1;
  gc.ec[tnXUNUX].initSpeed = 100;
  gc.ec[tnXUNUX].speedBonus = 5;
  gc.ec[tnBOMBER].ammoPerShoot = 10;
  gc.ec[tnUTETRIS].speedBonus = 1;
  
  gc.ballSpeed = 7;
  
  gc.bombInitRange = 1;
  gc.bombBonusRange = 1;
  
  gc.castleInitEnergy = 1000;
  gc.castleRotSpeed = 10;
  gc.castleRegenSpeed = 25;
  gc.castlePowerBase = 500;
  gc.ammoPerCannon = 50;
  
  gc.utetrisMinLength = 15;

  gc.initFuel = 5000;
  gc.bonusFuel = 1000;
  gc.thrustAcc = 50;
  gc.fuelUseSpeed = 20;
  
  gc.wormFlySpeed = 12;
  gc.wormWalkSpeed = 24;
  gc.ninjaRopeLength = 40;
  gc.ninjaPullSpeed = 18;

  gc.ballsTakeBonuses = true;
  gc.regenerateLevel = gc.regenerateScore = gc.regenerateAmmo = gc.regenerateDestroyed = true;
  
  gc.bulletsKillBalls = gc.cannonsKillBalls = gc.explosionsDestroyBalls = false;
  
  gc.ballsDestroyXunux = gc.ballsDestroyWalls = true;
  
  gc.xunuxKillsSnake = gc.xunuxKillsThruster = gc.xunuxKillsXunux = false;
  
  gc.xunuxAttacksXunux = true;
  }

FILE *cfgfile;

bool saving;

string currmsg, currval;

int lpid = 0;  

void save(char *msg, int& i) {
  if(saving) fprintf(cfgfile, "%s=%d\n", msg, i);
  else if(currmsg == msg) i = atoi(currval.c_str());
  }

void save(char *msg, int32& i) {
  if(saving) fprintf(cfgfile, "%s=%d\n", msg, i);
  else if(currmsg == msg) i = atoi(currval.c_str());
  }

void save(char *msg, TEntityConfig& ec) {
  if(saving) 
    fprintf(cfgfile, "%s=%d %d %d %d %d %d\n", msg, ec.initSpeed, ec.speedBonus, ec.maxSpeed, ec.initAmmo, ec.bonusAmmo, ec.ammoPerShoot);
  else if(currmsg == msg) 
    sscanf(currval.c_str(), "%d%d%d%d%d%d", &ec.initSpeed, &ec.speedBonus, &ec.maxSpeed, &ec.initAmmo, &ec.bonusAmmo, &ec.ammoPerShoot);
  }

void save(char *msg, bool& b) {
  if(saving) fprintf(cfgfile, "%s=%s\n", msg, b ? "yes" : "no");
  else if(currmsg == msg) b = currval.size() && (currval[0] & 1);
  }

void saveHex(char *msg, int& i) {
  if(saving) fprintf(cfgfile, "%s=%06x\n", msg, i);
  else if(currmsg == msg) sscanf(currval.c_str(), "%x", &i);
  }

void save(char *msg, string& s) {
  if(saving) fprintf(cfgfile, "%s=%s\n", msg, s.c_str());
  else if(currmsg == msg) s = currval;
  }

void eline() { if(saving) fprintf(cfgfile, "\n"); }

void cmt(char *cmt) { if(saving) fprintf(cfgfile, "# %s\n", cmt); }

void saveloadConfig() {
  cmt("Untahris Common Playground 3.1 configuration file");
  eline();
  
  cmt("Network settings");
  save("gametype", (int&) c.gt);
  save("server", c.server);
  save("port", c.port);
  eline();
  
  cmt("Screen settings");
  save("resx", c.resX);
  save("resy", c.resY);
  save("fullscreen", c.fs);
  save("resolution", c.res);
  save("pctx", c.pctX);
  save("pcty", c.pctY);
  save("scorelines", c.scorelines);
  eline();
  
  cmt("Map settings");
  save("initial world", c.initialWorld);
  save("mapname", c.mapname);
  save("initial bonuses", c.initBonus);
  save("background", c.background);
  save("graphics type", c.graphicsType);
  eline();
  
  cmt("For each player type: init speed, speed bonus, max speed bonuses, init ammo, bonus ammo, ammo per shoot");
  for(int i=0; i<9; i++)
    save(charnames[i], gc.ec[i]);
  eline();
  
  cmt("Other game variables");
  save("thruster initFuel", gc.initFuel);
  save("thruster bonusFuel", gc.bonusFuel);
  save("thruster acceleration", gc.thrustAcc);
  save("thruster fuel usage", gc.fuelUseSpeed);
  save("bomber initRange", gc.bombInitRange);
  save("bomber bonusRange", gc.bombBonusRange);
  save("utetris minLength", gc.utetrisMinLength);
  save("castle initEnergy", gc.castleInitEnergy);
  save("castle rotSpeed", gc.castleRotSpeed);
  save("castle regeneration", gc.castleRegenSpeed);
  save("castle power per bonus", gc.castlePowerBase);
  save("castle ammo per cannon", gc.ammoPerCannon);
  save("worm fly speed", gc.wormFlySpeed);
  save("worm walk speed", gc.wormWalkSpeed);
  save("worm ninja rope length", gc.ninjaRopeLength);
  save("worm ninja rope pull speed", gc.ninjaPullSpeed);
  
  save("ball speed", gc.ballSpeed);  

  save("regenerate level", gc.regenerateLevel);
  save("regenerate score", gc.regenerateScore);
  save("regenerate ammo", gc.regenerateAmmo);
  save("regenerate destroyed by bombs", gc.regenerateDestroyed);
  save("balls killed by bullets", gc.bulletsKillBalls);
  save("balls killed by cannons", gc.cannonsKillBalls);
  save("balls killed by explosions", gc.explosionsDestroyBalls);
  save("balls destroy xunux walls", gc.ballsDestroyXunux);
  save("balls destroy other walls", gc.ballsDestroyWalls);
  save("balls take bonuses", gc.ballsTakeBonuses);
  
  save("unstable xunux kills snake/tron", gc.xunuxKillsSnake);
  save("unstable xunux kills thruster", gc.xunuxKillsThruster);
  save("unstable xunux kills xunux", gc.xunuxKillsXunux);
  save("xunux may attack xunux", gc.xunuxAttacksXunux);
  eline();
    
  for(int i=0; i<size(lps); i++) {
    if(saving) lpid = i;
    
    lpid++;
    save("player", lpid);
    if(!saving && currmsg == "player") lps.resize(lpid);
    lpid--;

    saveHex("color", lps[lpid].color);
    save("keyUp", lps[lpid].keys[0]);
    save("keyRight", lps[lpid].keys[1]);
    save("keyDown", lps[lpid].keys[2]);
    save("keyLeft", lps[lpid].keys[3]);
    save("keyFire", lps[lpid].keys[4]);
    save("keyEarth", lps[lpid].keys[5]);
    save("keyAir", lps[lpid].keys[6]);
    save("keyWater", lps[lpid].keys[7]);
    eline();
    
    if(!saving) break;
    }
  }

void saveConfig() {
  cfgfile = fopen(configfile, "wt");
  if(!cfgfile) return;
  saving = true;
  saveloadConfig();
  fclose(cfgfile);
  }

void loadConfig() {
  defaultConfig();
  FILE *cfgfile = fopen(configfile, "rt");
  if(!cfgfile) return;
  saving = false;
  
  string s;
  while((s = readStr(cfgfile)) != "") {
    if(s[0] == '#') continue;
    int pos = s.find("=");
    currmsg = s.substr(0, pos);
    currval = s.substr(pos+1);
    
    saveloadConfig();
    }
  fclose(cfgfile);
  }

int main(int argc, char **argv) {
  g.sizeX = 0;
  g.sizeY = 0;
  d.ofsX = 0;
  d.ofsY = 0;
  d.Res = 16;
  if(argc == 2) configfile = argv[1];
  initDegs();

  printf("Untahris (Common Playground) v3.1\n");
  printf("copyright (C) 2006, 2007, 2009 longuens@users.sourceforge.net\n");
  printf("Released under GNU General Public License (see the file COPYING)\n");

  try { 
    srand(time(NULL));
    loadConfig();
  
    d.init();
  
    printf("Running.\n");
    checkConfig();
    d.menuLoop();
  
    saveConfig();
    }
  catch(TException e) {
    printf("Exception catched\n");
    printf("%s/%d/%s\n", e.msg.c_str(), e.data, e.ds.c_str());
    }
  }
